#include "devil.h"

int setup_devil() {
  ilInit();
  ilutRenderer(ILUT_OPENGL);
  return 0;
}

/* Load an image into a GL texture */
ILuint load_gl_image(ILuint *image, char *filename) {
  ILuint error;

  /* Generate image */
  ilGenImages(1, image);
  error = ilGetError();
  if (error) return error;

  /* Bind this image */
  ilBindImage(*image);

  /* Load file */
  if (! ilLoadImage(filename)) return 1;

  /* Worked, make into OpenGL image */
  *image = ilutGLBindTexImage();
  error = ilGetError();
  if (error) return error;

  /* Make mipmaps */
  ilutGLBuildMipmaps();

  return 0;
}

/*
  Try to load an image first in cwd then in dirname.
  On success, returns 0 and puts the OpenGL texture ID in texture_id.
  On failure, returns non-zero (*texture_id undefined in this case).
*/
int load_image(char *dirname, char *filename, GLuint *texture_id) {

  /* Try local */
  FILE *file = open_read(filename);
  if (file) {
    close_file(file);
    if (load_gl_image(texture_id, filename)) return 1;
    printf("Loaded %s\n", filename);
    return 0;
  }

  /* No good */
  char actual[UL_NAME_MAX + 1];
  if (find_file(dirname, filename, actual)) return 2;

  if (load_gl_image(texture_id, actual)) return 3;
  printf("Loaded %s\n", actual);
  return 0;
}

int take_screenshot(char *filename) {
  ILuint image;

  ilGenImages(1, &image);
  if (ilGetError()) return 1;
  ilBindImage(image);

  if (! ilutGLScreen()) return 2;

  ilEnable(IL_FILE_OVERWRITE);
  ilSaveImage(filename);
  if (ilGetError()) return 4;

  return 0;
}
