#include "ms3d.h"

/* Read a model */
int read_ms3d(ms3d_model_t &ms3d, char *filename) {
  unsigned short i, j;
  FILE *file = open_read(filename);
  if (! file) return 1;

  /* Read header */
  ms3d_header_t &header = ms3d.header;
  if (read_arbitrary(file, &header.magic, sizeof(header.magic))) return 2;
  if (read_uint32(file, &header.version)) return 3;
  if (strncmp(header.magic, MS3D_MAGIC, sizeof(header.magic))) return -1;
  if (header.version > MS3D_VERSION) return -2;

  /* Read number of vertices */
  if (read_uint16(file, &ms3d.num_vertices)) return 4;
  ms3d.vertices = new ms3d_vertex_t[ms3d.num_vertices];

  /* Read vertices */
  for (i = 0; i < ms3d.num_vertices; i++) {
    ms3d_vertex_t &vertex = ms3d.vertices[i];
    /* Flags */
    if (read_uint8(file, &vertex.flags)) return 5;
    /* Co-ordinates */
    if (read_vec3(file, &vertex.position)) return 6;
    /* Bone ID */
    if (read_int8(file, &vertex.bone_id)) return 7;
    /* Reference count */
    if (read_uint8(file, &vertex.reference_count)) return 8;
  }

  /* Read number of triangles */
  if (read_uint16(file, &ms3d.num_triangles)) return 9;
  ms3d.triangles = new ms3d_triangle_t[ms3d.num_triangles];

  /* Read triangles */
  for (i = 0; i < ms3d.num_triangles; i++) {
    ms3d_triangle_t &triangle = ms3d.triangles[i];
    /* Flags */
    if (read_uint16(file, &triangle.flags)) return 10;
    /* Vertices */
    if (read_uint16(file, &triangle.vertex_id[0])) return 11;
    if (read_uint16(file, &triangle.vertex_id[1])) return 12;
    if (read_uint16(file, &triangle.vertex_id[2])) return 13;
    /* Normals */
    if (read_vec3(file, &triangle.normals[0])) return 14;
    if (read_vec3(file, &triangle.normals[1])) return 15;
    if (read_vec3(file, &triangle.normals[2])) return 16;
    /* UVs */
    if (read_float(file, &triangle.s[0])) return 17;
    if (read_float(file, &triangle.s[1])) return 18;
    if (read_float(file, &triangle.s[2])) return 19;
    if (read_float(file, &triangle.t[0])) return 20;
    if (read_float(file, &triangle.t[1])) return 21;
    if (read_float(file, &triangle.t[2])) return 22;
    /* Smoothing group */
    if (read_uint8(file, &triangle.smoothing_group)) return 23;
    /* Group ID */
    if (read_uint8(file, &triangle.group_id)) return 24;
  }

  /* Read number of groups */
  if (read_uint16(file, &ms3d.num_groups)) return 25;
  ms3d.groups = new ms3d_group_t[ms3d.num_groups];

  /* Read groups */
  for (i = 0; i < ms3d.num_groups; i++) {
    ms3d_group_t &group = ms3d.groups[i];
    /* Flags */
    if (read_uint8(file, &group.flags)) return 26;
    /* Name */
    if (read_arbitrary(file, group.name, sizeof(group.name))) return 27;
    /* Number of triangles */
    if (read_uint16(file, &group.num_triangles)) return 28;
    group.triangle_id = new unsigned short[group.num_triangles];
    for (j = 0; j < group.num_triangles; j++) {
      if (read_uint16(file, &group.triangle_id[j])) return 29;
    }
    /* Material ID */
    if (read_int8(file, &group.material_id)) return 30;
  }

  /* Read number of materials */
  if (read_uint16(file, &ms3d.num_materials)) return 31;
  ms3d.materials = new ms3d_material_t[ms3d.num_materials];

  /* Read materials */
  for (i = 0; i < ms3d.num_materials; i++) {
    ms3d_material_t &material = ms3d.materials[i];
    /* Name */
    if (read_arbitrary(file, material.name, sizeof(material.name))) return 32;
    /* Ambient */
    for (j = 0; j < 4; j++) {
      if (read_float(file, &material.ambient[j])) return 33;
    }
    /* Diffuse */
    for (j = 0; j < 4; j++) {
      if (read_float(file, &material.diffuse[j])) return 34;
    }
    /* Specular */
    for (j = 0; j < 4; j++) {
      if (read_float(file, &material.specular[j])) return 35;
    }
    /* Emissive */
    for (j = 0; j < 4; j++) {
      if (read_float(file, &material.emissive[j])) return 36;
    }
    /* Shininess */
    if (read_float(file, &material.shininess)) return 37;
    /* Opacity */
    if (read_float(file, &material.opacity)) return 38;
    /* Mode */
    if (read_int8(file, &material.mode)) return 39;
    /* Texture */
    if (read_arbitrary(file, material.texture, sizeof(material.texture))) return 40;
    /* Alpha */
    if (read_arbitrary(file, material.alphamap, sizeof(material.alphamap))) return 41;
  }

  /* Read animation stuff */
  if (read_float(file, &ms3d.framerate)) return 42;
  if (read_float(file, &ms3d.current_time)) return 43;
  if (read_uint32(file, &ms3d.num_frames)) return 44;

  /* Read number of bones */
  if (read_uint16(file, &ms3d.num_joints)) return 45;
  ms3d.joints = new ms3d_joint_t[ms3d.num_joints];

  /* Read bones */
  for (i = 0; i < ms3d.num_joints; i++) {
    ms3d_joint_t &joint = ms3d.joints[i];
    /* Flags */
    if (read_uint8(file, &joint.flags)) return 46;
    /* Name */
    if (read_arbitrary(file, joint.name, sizeof(joint.name))) return 47;
    /* Parent */
    if (read_arbitrary(file, joint.parent_name, sizeof(joint.parent_name))) return 48;
    /* Rotation */
    if (read_vec3(file, &joint.rotation)) return 49;
    /* Position */
    if (read_vec3(file, &joint.position)) return 50;
    /* Number of rotation keyframes */
    if (read_uint16(file, &joint.num_rotation_keyframes)) return 51;
    /* Number of position keyframes */
    if (read_uint16(file, &joint.num_position_keyframes)) return 52;
    joint.rotation_keyframes = new ms3d_keyframe_rot_t[joint.num_rotation_keyframes];
    joint.position_keyframes = new ms3d_keyframe_pos_t[joint.num_position_keyframes];
    /* Rotation keyframes */
    for (j = 0; j < joint.num_rotation_keyframes; j++) {
      ms3d_keyframe_rot_t &keyframe = joint.rotation_keyframes[j];
      /* Time */
      if (read_float(file, &keyframe.time)) return 53;
      /* Rotation */
      if (read_vec3(file, &keyframe.rotation)) return 54;
    }
    /* Position keyframes */
    for (j = 0; j < joint.num_position_keyframes; j++) {
      ms3d_keyframe_pos_t &keyframe = joint.position_keyframes[j];
      /* Time */
      if (read_float(file, &keyframe.time)) return 55;
      /* Rotation */
      if (read_vec3(file, &keyframe.position)) return 56;
    }
  }

  close_file(file);

  return 0;
}

/* Write the model */
int write_ms3d(ms3d_model_t &ms3d, char *filename) {
  unsigned short i, j;
  FILE *file = open_write(filename);
  if (! file) return 1;

  /* Write header */
  ms3d_header_t &header = ms3d.header;
  if (write_arbitrary(file, &header.magic, sizeof(header.magic))) return 2;
  if (write_uint32(file, &header.version)) return 3;

  /* Write number of vertices */
  if (write_uint16(file, &ms3d.num_vertices)) return 4;

  /* Write vertices */
  for (i = 0; i < ms3d.num_vertices; i++) {
    ms3d_vertex_t &vertex = ms3d.vertices[i];
    /* Flags */
    if (write_char(file, (char *) &vertex.flags)) return 5;
    /* Co-ordinates */
    if (write_vec3(file, &vertex.position)) return 6;
    /* Bone ID */
    if (write_char(file, &vertex.bone_id)) return 7;
    /* Reference count */
    if (write_char(file, (char *) &vertex.reference_count)) return 8;
  }

  /* Write number of triangles */
  if (write_uint16(file, &ms3d.num_triangles)) return 9;

  /* Write triangles */
  for (i = 0; i < ms3d.num_triangles; i++) {
    ms3d_triangle_t &triangle = ms3d.triangles[i];
    /* Flags */
    if (write_uint16(file, &triangle.flags)) return 10;
    /* Vertices */
    if (write_uint16(file, &triangle.vertex_id[0])) return 11;
    if (write_uint16(file, &triangle.vertex_id[1])) return 12;
    if (write_uint16(file, &triangle.vertex_id[2])) return 13;
    /* Normals */
    if (write_vec3(file, &triangle.normals[0])) return 14;
    if (write_vec3(file, &triangle.normals[1])) return 15;
    if (write_vec3(file, &triangle.normals[2])) return 16;
    /* UVs */
    if (write_float(file, &triangle.s[0])) return 17;
    if (write_float(file, &triangle.s[1])) return 18;
    if (write_float(file, &triangle.s[2])) return 19;
    if (write_float(file, &triangle.t[0])) return 20;
    if (write_float(file, &triangle.t[1])) return 21;
    if (write_float(file, &triangle.t[2])) return 22;
    /* Smoothing group */
    if (write_char(file, (char *) &triangle.smoothing_group)) return 23;
    /* Group ID */
    if (write_char(file, (char *) &triangle.group_id)) return 24;
  }

  /* Write number of groups */
  if (write_uint16(file, &ms3d.num_groups)) return 25;

  /* Write groups */
  for (i = 0; i < ms3d.num_groups; i++) {
    ms3d_group_t &group = ms3d.groups[i];
    /* Flags */
    if (write_char(file, (char *) &group.flags)) return 26;
    /* Name */
    if (write_arbitrary(file, group.name, sizeof(group.name))) return 27;
    /* Number of triangles */
    if (write_uint16(file, &group.num_triangles)) return 28;
    for (j = 0; j < group.num_triangles; j++) {
      if (write_uint16(file, &group.triangle_id[j])) return 29;
    }
    /* Material ID */
    if (write_char(file, &group.material_id)) return 30;
  }

  /* Write number of materials */
  if (write_uint16(file, &ms3d.num_materials)) return 31;

  /* Write materials */
  for (i = 0; i < ms3d.num_materials; i++) {
    ms3d_material_t &material = ms3d.materials[i];
    /* Name */
    if (write_arbitrary(file, material.name, sizeof(material.name))) return 32;
    /* Ambient */
    for (j = 0; j < 4; j++) {
      if (write_float(file, &material.ambient[j])) return 33;
    }
    /* Diffuse */
    for (j = 0; j < 4; j++) {
      if (write_float(file, &material.diffuse[j])) return 34;
    }
    /* Specular */
    for (j = 0; j < 4; j++) {
      if (write_float(file, &material.specular[j])) return 35;
    }
    /* Emissive */
    for (j = 0; j < 4; j++) {
      if (write_float(file, &material.emissive[j])) return 36;
    }
    /* Shininess */
    if (write_float(file, &material.shininess)) return 37;
    /* Opacity */
    if (write_float(file, &material.opacity)) return 38;
    /* Mode */
    if (write_char(file, &material.mode)) return 39;
    /* Texture */
    if (write_arbitrary(file, material.texture, sizeof(material.texture))) return 40;
    /* Alpha */
    if (write_arbitrary(file, material.alphamap, sizeof(material.alphamap))) return 41;
  }

  /* Write animation stuff */
  if (write_float(file, &ms3d.framerate)) return 42;
  if (write_float(file, &ms3d.current_time)) return 43;
  if (write_uint32(file, &ms3d.num_frames)) return 44;

  /* Write number of bones */
  if (write_uint16(file, &ms3d.num_joints)) return 45;

  /* Write bones */
  for (i = 0; i < ms3d.num_joints; i++) {
    ms3d_joint_t &joint = ms3d.joints[i];
    /* Flags */
    if (write_char(file, (char *) &joint.flags)) return 46;
    /* Name */
    if (write_arbitrary(file, joint.name, sizeof(joint.name))) return 47;
    /* Parent */
    if (write_arbitrary(file, joint.parent_name, sizeof(joint.parent_name))) return 48;
    /* Rotation */
    if (write_vec3(file, &joint.rotation)) return 49;
    /* Position */
    if (write_vec3(file, &joint.position)) return 50;
    /* Number of rotation keyframes */
    if (write_uint16(file, &joint.num_rotation_keyframes)) return 51;
    /* Number of position keyframes */
    if (write_uint16(file, &joint.num_position_keyframes)) return 52;
    /* Rotation keyframes */
    for (j = 0; j < joint.num_rotation_keyframes; j++) {
      ms3d_keyframe_rot_t &keyframe = joint.rotation_keyframes[j];
      /* Time */
      if (write_float(file, &keyframe.time)) return 53;
      /* Rotation */
      if (write_vec3(file, &keyframe.rotation)) return 54;
    }
    /* Position keyframes */
    for (j = 0; j < joint.num_position_keyframes; j++) {
      ms3d_keyframe_pos_t &keyframe = joint.position_keyframes[j];
      /* Time */
      if (write_float(file, &keyframe.time)) return 55;
      /* Rotation */
      if (write_vec3(file, &keyframe.position)) return 56;
    }
  }

  close_file(file);

  return 0;
}

/* Describe this model */
void describe_ms3d(ms3d_model_t &ms3d, unsigned int flags) {
  unsigned short i, j;

  if (! flags) return;

  /* Describe header */
  if (flags & MS3D_DESCRIBE_HEADER) {
    printf("MS3D magic '%10s' (version %u)\n\n", ms3d.header.magic, ms3d.header.version);
  }

  /* Describe vertices */
  if (flags & MS3D_DESCRIBE_VERTICES) {
    printf("MS3D vertices: %u\n", ms3d.num_vertices);
    for (i = 0; i < ms3d.num_vertices; i++) {
      ms3d_vertex_t &vertex = ms3d.vertices[i];
      printf("  Vertex % 4u: (% .4f, % .4f, % .4f) [bone %d]\n", i, vertex.position[0], vertex.position[1], vertex.position[2], (int) vertex.bone_id);
    }
    printf("\n");
  }

  /* Describe triangles */
  if (flags & MS3D_DESCRIBE_TRIANGLES) {
    printf("Model triangles: %u\n", ms3d.num_triangles);
    for (i = 0; i < ms3d.num_triangles; i++) {
      ms3d_triangle_t &triangle = ms3d.triangles[i];
      /* Trying to fit this into DOS box */
      printf("  Tri % 4u: (%u, %u, %u) [UV (%0.2f,%0.2f),(%0.2f,%0.2f),(%0.2f,%0.2f)] [grp %u]\n", i, triangle.vertex_id[0], triangle.vertex_id[1], triangle.vertex_id[2], triangle.s[0], triangle.t[0], triangle.s[1], triangle.t[1], triangle.s[2], triangle.t[2], (unsigned int) triangle.group_id);
    }
    printf("\n");
  }

  /* Describe groups */
  if (flags & MS3D_DESCRIBE_GROUPS) {
    printf("Model groups: %u\n", ms3d.num_groups);
    for (i = 0; i < ms3d.num_groups; i++) {
      ms3d_group_t &group = ms3d.groups[i];
      printf("  Group % 2u: '%s' [material %d]\n", i, group.name, (int) group.material_id);

      /* Triangles in group */
      if (group.num_triangles) {
        printf("    Triangles: %u\n     ", group.num_triangles);
        for (j = 0; j < group.num_triangles; j++) {
          printf(" %  4u", group.triangle_id[j]);
          /* Arrange in columns */
          if ((j + 1) % 13) continue;
          /* Will be a newline after last row anyway */
          if (j == group.num_triangles - 1) continue;
          printf("\n     ");
        }
        printf("\n");
      }
    }
    printf("\n");
  }

  /* Describe materials */
  if (flags & MS3D_DESCRIBE_MATERIALS) {
    printf("Model materials: %u\n", ms3d.num_materials);
    for (i = 0; i < ms3d.num_materials; i++) {
      ms3d_material_t &material = ms3d.materials[i];
      printf("  Mat % 2u: '%s' [texture '%s']\n", i, material.name, material.texture);
    }
    printf("\n");
  }

  /* Describe joints */
  if (flags & MS3D_DESCRIBE_JOINTS) {
    printf("Model joints: %u\n", ms3d.num_joints);
    for (i = 0; i < ms3d.num_joints; i++) {
      ms3d_joint_t &joint = ms3d.joints[i];
      printf("  Joint % 2d: '%s' [parent '%s']\n", i, joint.name, joint.parent_name);
      printf("    Position (% .4f, % .4f, %.4f) Rotation (% .2f, % .2f, %.2f)\n", joint.position[0], joint.position[1], joint.position[2], SG_RADIANS_TO_DEGREES * joint.rotation[0], SG_RADIANS_TO_DEGREES * joint.rotation[1], SG_RADIANS_TO_DEGREES * joint.rotation[2]);

      /* Describe animation */
      if (flags & MS3D_DESCRIBE_ANIMATION) {
        /* Describe position frames */
        printf("    Position keyframes: %u\n", joint.num_position_keyframes);
        for (j = 0; j < joint.num_position_keyframes; j++) {
          ms3d_keyframe_pos_t &keyframe = joint.position_keyframes[j];
          printf("    Position keyframe % 2u: (% .4f, % .4f, % .4f)\n", j, keyframe.position[0], keyframe.position[1], keyframe.position[2]);
        }

        /* Describe rotation frames */
        printf("    Rotation keyframes: %u\n", joint.num_rotation_keyframes);
        for (j = 0; j < joint.num_position_keyframes; j++) {
          ms3d_keyframe_rot_t &keyframe = joint.rotation_keyframes[j];
          printf("    Rotation keyframe % 2u: (% .2f, % .2f, % .2f)\n", j, SG_RADIANS_TO_DEGREES * keyframe.rotation[0], SG_RADIANS_TO_DEGREES * keyframe.rotation[1], SG_RADIANS_TO_DEGREES * keyframe.rotation[2]);
        }
      }
    }
    printf("\n");
  }
}

/* Free memory used by model */
void cleanup_ms3d(ms3d_model_t *ms3d) {
  unsigned short i;

  if (! ms3d) return;

  /* Meshes */
  if (ms3d->num_groups && ms3d->groups) delete[] ms3d->groups;
  if (ms3d->num_vertices && ms3d->vertices) delete[] ms3d->vertices;
  if (ms3d->num_triangles && ms3d->triangles) delete[] ms3d->triangles;

  /* Materials */
  if (ms3d->num_materials && ms3d->materials) delete[] ms3d->materials;

  /* Joints */
  for (i = 0; i < ms3d->num_joints; i++) {
    ms3d_joint_t *joint = &ms3d->joints[i];
    if (joint->num_rotation_keyframes && joint->rotation_keyframes) delete[] joint->rotation_keyframes;
    if (joint->num_position_keyframes && joint->position_keyframes) delete[] joint->position_keyframes;
  }
  if (ms3d->num_joints && ms3d->joints) delete[] ms3d->joints;

  delete ms3d;
}
