/* Milkshape .ms3d file stuff */
#ifndef MS3D_H
#define MS3D_H

#include <plib/sg.h>
#include <stdio.h>
#include "file.h"

#define MS3D_MAGIC "MS3D000000"
#define MS3D_VERSION 4

#define MS3D_NAME_LENGTH 32
#define MS3D_TEXTURE_LENGTH 128

#define MS3D_FLAG_SELECTED 1
#define MS3D_FLAG_HIDDEN 2
#define MS3D_FLAG_SELECTED2 4
#define MS3D_FLAG_DIRTY 8

typedef struct {
  char magic[10];
  unsigned int version;
} ms3d_header_t;

typedef struct {
  unsigned char flags; /* SELECTED | SELECTED2 | HIDDEN */
  sgVec3 position;
  char bone_id;
  unsigned char reference_count; /* number of triangles containing the vertex */
} ms3d_vertex_t;

typedef struct {
  unsigned short flags; /* SELECTED | SELECTED2 | HIDDEN */
  unsigned short vertex_id[3];
  sgVec3 normals[3];
  float s[3]; /* U */
  float t[3]; /* V */
  unsigned char smoothing_group;
  unsigned char group_id;
} ms3d_triangle_t;

typedef struct {
  unsigned char flags; /* SELECTED | HIDDEN */
  char name[MS3D_NAME_LENGTH];
  unsigned short num_triangles;
  unsigned short *triangle_id;
  char material_id; /* -1 => no material */
} ms3d_group_t;

typedef struct {
  char name[MS3D_NAME_LENGTH];
  float ambient[4];
  float diffuse[4];
  float specular[4];
  float emissive[4];
  float shininess; /* 0.0 - 128.0 */
  float opacity; /* ms3d spec calls it transparency, this is wrong; 0.0 - 1.0 */
  char mode; /* ms3d spec says "0, 1, 2 is unused now" */
  char texture[MS3D_TEXTURE_LENGTH];
  char alphamap[MS3D_TEXTURE_LENGTH];
} ms3d_material_t;

typedef struct {
  float time;
  sgVec3 rotation; /* ms3d spec uses float[3] */
} ms3d_keyframe_rot_t;

typedef struct {
  float time;
  sgVec3 position;
} ms3d_keyframe_pos_t;

typedef struct {
  unsigned char flags; /* SELECTED | DIRTY */
  char name[MS3D_NAME_LENGTH];
  char parent_name[MS3D_NAME_LENGTH];
  sgVec3 rotation;
  sgVec3 position;
  unsigned short num_rotation_keyframes;
  unsigned short num_position_keyframes;
  ms3d_keyframe_rot_t *rotation_keyframes;
  ms3d_keyframe_pos_t *position_keyframes;
} ms3d_joint_t;

typedef struct {
  ms3d_header_t header;
  unsigned short num_vertices;
  unsigned short num_triangles;
  unsigned short num_groups;
  unsigned short num_materials;
  float framerate;
  float current_time; /* irrelevant */
  unsigned int num_frames;
  unsigned short num_joints;
  ms3d_vertex_t *vertices;
  ms3d_triangle_t *triangles;
  ms3d_group_t *groups;
  ms3d_material_t *materials;
  ms3d_joint_t *joints;
} ms3d_model_t;

#define MS3D_DESCRIBE_HEADER    0x0001
#define MS3D_DESCRIBE_VERTICES  0x0002
#define MS3D_DESCRIBE_TRIANGLES 0x0004
#define MS3D_DESCRIBE_GROUPS    0x0008
#define MS3D_DESCRIBE_MATERIALS 0x0010
#define MS3D_DESCRIBE_JOINTS    0x0020
#define MS3D_DESCRIBE_ANIMATION 0x0040
#define MS3D_DESCRIBE_ALL       0x007F

int read_ms3d(ms3d_model_t &, char *);
int write_ms3d(ms3d_model_t &, char *);
void describe_ms3d(ms3d_model_t &, unsigned int);
void cleanup_ms3d(ms3d_model_t *);

#endif
